% Multi-task GGMs
%   [X_allvars obj_func] = learn_multitask_ggm(S_allvars, NS, rho, rho_diag, group_norm, T);
% Input:
%   S_allvars: sample covariance per task (#vars x #vars x #tasks array)
%   NS: number of samples per task (#tasks vector)
%   rho: regularization parameter for encouraging sparseness
%   group_norm: 2 for L12, inf for L1inf
%   T: number of iterations (10 is usually enough)
% Output:
%   X_allvars: precision matrix per task (#vars x #vars x #tasks array)
%   obj_func: objective function value after each iteration
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function [X_allvars obj_func] = learn_multitask_ggm(S_allvars, NS, rho, rho_diag, group_norm, T);

% Check input
if group_norm ~= 2 & group_norm ~= inf
    error('group_norm must be either 2 or inf');
end
C = length(NS);
if size(S_allvars,3) ~= length(NS)
    error('S_allvars and NS should have the same number of tasks');
end
NS = NS(:) / sum(NS);

% Remove variables from optimization
N = size(S_allvars,1);
norms = zeros(N,N);
if group_norm == 2
    for c = 1:C
        norms = norms + NS(c)^2*(S_allvars(:,:,c).^2);
    end
    threshold = rho^2;
else
    for c = 1:C
        norms = norms + NS(c)*abs(S_allvars(:,:,c));
    end
    threshold = rho;
end
norms = norms - diag(diag(norms));
norms = max(norms,[],1);
idx_zero = find(norms <= threshold);
idx_opti = setdiff(1:N,idx_zero);
N = length(idx_opti);
if N == 1
    idx_zero = 1:size(S_allvars,1);
    idx_opti = [];
    N = 0;
end
fprintf('%d variables for optimization, %d variables removed...\n',N,length(idx_zero));
S = S_allvars(idx_opti,idx_opti,:);
for c = 1:C
    X(:,:,c) = diag(1 ./ diag(S(:,:,c)));
end

% Iterate T times
obj_func = [];
if N > 0
    first = true;
    for t = 1:T
        fprintf('Iteration %d variable ', t);
        for n = 1:N
            if n > 1
                for rep = 1:fix(log10(n-1)+1)
                    fprintf('\b');
                end
            end
            fprintf('%d', n);
            idx = [1:n-1 n+1:N];
            W = X(idx,idx,:);
            u = squeeze(S(idx,n,:));
            if N == 2
                u = u.'; % solving squeeze problem
            end
            v = squeeze(S(n,n,:));
            y = squeeze(X(idx,n,:));
            if N == 2
                y = y.'; % solving squeeze problem
            end
            z = squeeze(X(n,n,:));
            if first

                % At the beginning, H is inverse of diagonal terms of W
                H = zeros(N-1,N-1,C);
                for c = 1:C
                    H(:,:,c) = diag(1 ./ diag(W(:,:,c)));
                end
                first = false;
            else

                % Compute inverse of W by Woodbury formula twice in order
                % to update previous inverse H
                if n == 1
                    idx_prev = [2:N-1 1];
                    H_prev = H_prev(idx_prev,idx_prev,:);
                    y_prev = y_prev(idx_prev,:);
                end
                for c = 1:C
                    delta = y_prev(:,c) - y(:,c);
                    if n == 1
                        pos = N-1;
                    else
                        pos = n-1;
                    end
                    delta(pos) = (z_prev(c) - z(c)) / 2;
                    H_prev_U = [H_prev(:,:,c) * delta, H_prev(:,pos,c)];
                    Vt_H_prev = [H_prev(pos,:,c); delta.' * H_prev(:,:,c)];
                    Vt_H_prev_U = [Vt_H_prev * delta, Vt_H_prev(:,pos)];
                    H(:,:,c) = H_prev(:,:,c) - H_prev_U * inv(eye(2) + Vt_H_prev_U) * Vt_H_prev;
                end
            end

            % Solve in closed form
            y = solve(NS,v,H,u,rho,group_norm,y);
            if rho_diag == 0
                for c = 1:C
                    z(c) = 1 / v(c) + y(:,c).' * H(:,:,c) * y(:,c);
                end
            else
                b = zeros(C,1);
                for c = 1:C
                    b(c) = y(:,c).' * H(:,:,c) * y(:,c);
                end
                z = solve_diag(NS,NS.*v,b,rho_diag,group_norm);
            end

            % Update matrices
            X(idx,n,:) = y;
            for c = 1:C
                X(n,idx,c) = y(:,c).';
            end
            X(n,n,:) = z;
            H_prev = H;
            y_prev = y;
            z_prev = z;
        end

        % Compute error
        E = 0;
        for c = 1:C
            [U p] = chol(X(:,:,c));
            if p ~= 0
                error('Precision matrix %d is not positive definite at iteration %d',c,t);
            end
            E = E + NS(c)*(2*sum(log(diag(U))) - sum(sum(S(:,:,c).*X(:,:,c))));
        end
        if group_norm == 2
            mt = sqrt(sum(X.^2,3));
        else
            mt = max(abs(X),[],3);
        end
        E = E - rho*sum(sum(mt - diag(diag(mt)))) - rho_diag*sum(diag(mt));
        fprintf(' obj.fun. %0.12f\n',E);
        obj_func = [obj_func E];
    end
end

% Return solution
N = size(S_allvars,1);
X_allvars = zeros(N,N,C);
X_allvars(idx_opti,idx_opti,:) = X;
for n = idx_zero
    if rho_diag == 0
        X_allvars(n,n,:) = 1 ./ S_allvars(n,n,:);
    else
        v = squeeze(S_allvars(n,n,:));
        b = zeros(C,1);
        X_allvars(n,n,:) = solve_diag(NS,NS.*v,b,rho_diag,group_norm);
    end
end

% Solve for one specific variable at a time
function y = solve(NS,v,H,u,rho,group_norm,y0);

N = size(u,1);
C = length(NS);
for n = 1:N
    idx = [1:n-1 n+1:N];
    q = NS .* v .* squeeze(H(n,n,:));
    HH = squeeze(H(idx,n,:));
    if N == 2
        HH = HH.'; % solving squeeze problem
    end
    Hy0 = sum(HH .* y0(idx,:),1);
    alpha = -NS .* (v .* Hy0.' + u(n,:).');
    if group_norm == 2

        % Solve separable quadratic trust region subproblem
        if sum(alpha.^2) <= rho*rho
            y0(n,:) = 0;
        else
            Tinner = 10;
            lam = 0;
            for tinner = 1:Tinner
                df = -sum((alpha.^2)./((1+lam*q).^2)) + rho*rho;
                hf = 2*sum(((alpha.^2).*q)./((1+lam*q).^3));
                lam_next = lam - df/hf;
                if lam_next < 0
                    fprintf('alpha = ');
                    disp(alpha.');
                    fprintf('q = ');
                    disp(q.');
                    error('This shouldn''t happen');
                end
                lam = lam_next;
            end
            y0(n,:) = lam*(alpha./(1+lam*q));
        end
    else
        
        % Solve continuous quadratic knapsack subproblem
        abs_alpha = abs(alpha);
        if sum(abs_alpha) <= rho
            y0(n,:) = 0;
        else
            [v_break idx] = sort(abs_alpha ./ q,'descend');
            sum_abs_alpha = cumsum(abs_alpha(idx));
            sum_q = cumsum(q(idx));
            c_sol = nan;
            for c = 1:C
                v_sol = (sum_abs_alpha(c) - rho)/sum_q(c);
                if c < C
                    v_break_c1 = v_break(c+1);
                else
                    v_break_c1 = 0;
                end
                if v_break(c) >= v_sol & v_sol >= v_break_c1
                    c_sol = c;
                    break
                end
            end
            if isnan(c_sol)
                fprintf('alpha = ');
                disp(alpha.');
                fprintf('q = ');
                disp(q.');
                error('This shouldn''t happen');
            end
            for c = 1:c_sol
                y0(n,idx(c)) = sign(alpha(idx(c))) * v_sol;
            end
            for c = c_sol+1:C
                y0(n,idx(c)) = alpha(idx(c)) / q(idx(c));
            end
        end
    end
end
y = y0;

% Solve for diagonal
function z = solve_diag(q,c,b,rho,group_norm);

if group_norm == 2

    % Solve separable logarithmic trust region subproblem
    Tinner = 10;
    lam = mean((q+b.*(c+rho))./(rho*(c+rho)));
    delta = lam/1000;
    for tinner = 1:Tinner
        f0 = f(lam-delta,q,c,b,rho);
        f1 = f(lam,q,c,b,rho);
        f2 = f(lam+delta,q,c,b,rho);
        df = (f2-f0)/(2*delta);
        hf = (f2-2*f1+f0)/delta^2;
        if abs(hf) < -1e-8
            % strange to happen, but what the heck
            break
        end
        lam_next = lam - df/hf;
        if lam_next < 0
            % just exit, use the previous lam
            break
        end
        lam = lam_next;
    end
    [tmp r] = f(lam,q,c,b,rho);
    z = b+q./(c+r);
else

    % Solve continuous logarithmic knapsack subproblem
    Tinner = 10;
    delta = 1e-3;
    K = length(q);
    min_v = max(b);
    v = q./c+b;
    [v_break idx] = sort(v,'descend');
    qs = q(idx);
    bs = b(idx);
    cs = c(idx);
    found = false;
    prev_oner = 0;
    for k = 1:K
        if k <= K-1 && v_break(k+1) > min_v
            if any(v_break(k+1) >= qs(1:k)./cs(1:k)+bs(1:k))
                error('Unexpected')
            end
            oner = sum(qs(1:k)./(v_break(k+1)-bs(1:k))-cs(1:k));
        else
            oner = inf;
        end
        if prev_oner <= rho & rho <= oner
            found = true;
            break
        end
        prev_oner = oner;
    end
    if ~found
        error('This should not happen');
    end
    if k == K
        v0 = min_v+delta;
    else
        v0 = v_break(k+1);
    end
    v = v0;
    for tinner = 1:Tinner
        if any(v >= qs(1:k)./cs(1:k)+bs(1:k))
            error('Unexpected')
        end
        f = sum(qs(1:k)./(v-bs(1:k))-cs(1:k)) - rho;
        df = sum(-qs(1:k)./(v-bs(1:k)).^2);
        if df > -1e-8
            % strange to happen, but what the heck
            break
        end
        v_next = v - f/df;
        if v_next < min_v
            % just exit, use the previous v
            break
        end
        v = v_next;
    end
    r = max(0,q./(v-b)-c);
    z = b+q./(c+r);
end

% For group_norm == 2
function [f_lam r] = f(lam,q,c,b,rho);

r = (b-lam*c + sqrt((b+lam*c).^2+4*lam*q)) / (2*lam);
f_lam = -sum(q.*log(r+c)) - b.'*r + lam/2*(r.'*r-rho^2);
