% Locally constant GGMs
%   [X_allvars errors] = learn_smooth_ggm(S_allvars, rho, rho_diag, tau, P, D_allvars, T);
% Input:
%   S_allvars: sample covariance
%   rho: regularization parameter for encouraging sparseness
%   tau: regularization parameter for encouraging flatness
%   P: dimensionality of space: 2 for images, 3 for volumes, etc.
%   D_allvars: matrix differential operator
%   T: number of iterations
% Output:
%   X: precision matrix per task (#vars x #vars x #tasks array)
%   obj_func: objective function value after each iteration
% Author: Jean Honorio
% E-mail: jhonorio@csail.mit.edu
function [X_allvars obj_func] = learn_smooth_ggm(S_allvars, rho, rho_diag, tau, P, D_allvars, T);

% Checks and inner constants
if tau > 0
    if size(D_allvars,2) ~= size(S_allvars,1)
        error('Wrong matrix differential operator size');
    end
end

% Remove variables from optimization
N = size(S_allvars,1);
S_nodiag = S_allvars-diag(diag(S_allvars));
norms = max(abs(S_nodiag),[],1);
threshold = rho - P * tau;
idx_zero = find(norms <= threshold);
idx_opti = setdiff(1:N,idx_zero);
N = length(idx_opti);
if N == 1
    idx_zero = 1:size(S_allvars,1);
    idx_opti = [];
    N = 0;
end
fprintf('%d variables for optimization, %d variables removed...\n',N,length(idx_zero));
S = S_allvars(idx_opti,idx_opti);
X = diag(1 ./ diag(S));
rhovec = rho*ones(N,1);
if tau > 0
    D = D_allvars(:,idx_opti);
    snz = sum(D~=0,2);
    if any(snz == 1)
        rhovec = rhovec + tau/2*sum(abs(D(snz == 1,:)),1).';
    end
    D = D(snz > 1,:);
else
    D = [];
end

% Iterate T times
obj_func = [];
if N > 0
    first = true;
    for t = 1:T
        fprintf('Iteration %d variable ', t);
        for n = 1:N
            if n > 1
                for rep = 1:fix(log10(n-1)+1)
                    fprintf('\b');
                end
            end
            fprintf('%d', n);
            idx = [1:n-1 n+1:N];
            W = X(idx,idx);
            u = S(idx,n);
            v = S(n,n);
            y = X(idx,n);
            z = X(n,n);
            if first

                % At the beginning, H is inverse of diagonal terms of W
                H = diag(1 ./ diag(W));
                first = false;
            else

                % Compute inverse of W by Woodbury formula twice in order
                % to update previous inverse H
                if n == 1
                    idx_prev = [2:N-1 1];
                    H_prev = H_prev(idx_prev,idx_prev);
                    y_prev = y_prev(idx_prev);
                end
                delta = y_prev - y;
                if n == 1
                    pos = N-1;
                else
                    pos = n-1;
                end
                delta(pos) = (z_prev - z) / 2;
                H_prev_U = [H_prev * delta, H_prev(:,pos)];
                Vt_H_prev = [H_prev(pos,:); delta.' * H_prev];
                Vt_H_prev_U = [Vt_H_prev * delta, Vt_H_prev(:,pos)];
                H = H_prev - H_prev_U * inv(eye(2) + Vt_H_prev_U) * Vt_H_prev;
            end

            % Solve in closed form for several directions 
            At = speye(N-1);
            b = zeros(N-1,1);
            f = rhovec(idx);
            dirs = speye(N-1);
            if tau > 0
                [At b f] = add_reg(At, b, f, N, n, idx, tau/2, D, W);
                kidx = find(D(:,n)==0).';
                dirs = [dirs, (D(kidx,idx)~=0).'];
            end
            y = solve(v + rho_diag, H, u, f, At.', b, y, dirs);
            z = (1 / (v + rho_diag)) + y.' * H * y;

            % Update matrices
            X(idx,n) = y;
            X(n,idx) = y.';
            X(n,n) = z;
            H_prev = H;
            y_prev = y;
            z_prev = z;
        end

        % Compute error
        [U p] = chol(X);
        if p ~= 0
            error('Matrix is not positive definite at iteration %d',t);
        end
        Xd = X;
        for n = 1:N
            Xd(n,n) = nan;
        end
        E = full(2*sum(log(diag(U)))) - sum(S(:).*X(:)) - rho*sum(abs(Xd(~isnan(Xd(:))))) - rho_diag*(sum(diag(X)));
        if tau > 0
            DXd = D*Xd;
            E = E - tau*sum(abs(DXd(~isnan(DXd(:)))));
        end
        fprintf(' error %0.12f\n',E);
        obj_func = [obj_func E];
    end
end

% Return solution
N = size(S_allvars,1);
X_allvars = zeros(N);
X_allvars(idx_opti,idx_opti) = X;
for n = idx_zero
    X_allvars(n,n) = 1 / (S_allvars(n,n) + rho_diag);
end

% Compute regularization projection
function [At b f] = add_reg(A0t, b0, f0, N, n, idx, fscalar, C, W);

% Add regularization across rows
kidx = find(C(:,n)==0).';
At = [A0t, C(kidx,idx).'];
b = [b0; zeros(length(kidx),1)];
f = [f0; fscalar * ones(length(kidx),1)];

% Add regularization across columns
midx = find(C(:,n)~=0).';
I = speye(N-1);
M = size(At,2);
for m = midx
    qm = C(m,idx);
    didx = find(qm == 0);
    Anewt = I(:,didx);
    bnew = (-1 / C(m,n)) * (W(didx,:) * qm.');
    fnew = abs(1 / C(m,n)) * fscalar * ones(length(didx),1);

    % Join with conditions A=I for b(.)=0, add new conditions for b(.)~=0
    for r = 1:length(didx)
        if bnew(r) ~= 0
            M = M+1;
            At(:,M) = Anewt(:,r);
            b(M) = bnew(r);
            f(M) = fnew(r);
        else
            f(didx(r)) = f(didx(r)) + fnew(r);
        end
    end
end

% Solve in closed form for several directions
function y = solve(v, Winv, u, f, A, b, y0, dirs);

% Iterate through all directions
Winvy0 = Winv * y0;
Ay0 = A * y0;
for i = 1:size(dirs,2)
    
    % Get the one-dimensional problem along direction d
    d = dirs(:,i);
    Winvd = Winv * d;
    Ad = A * d;
    p = v * (d.' * Winvd);
    q = (v * Winvy0 + u).' * d;
    t = Ad;
    r = f;
    s = -Ay0 + b;
    idx = find(t ~= 0);
    t = t(idx);
    r0 = r(idx) .* abs(t);
    s0 = s(idx) ./ t;

    % Sort discontinuity points and join the similar ones
    [tmp idx2] = sort(s0);
    M = 1;
    r = r0(idx2(1));
    s = s0(idx2(1));
    for m = 2:length(idx2)
        if s0(idx2(m)) == s0(idx2(m-1))
            r(M) = r(M) + r0(idx2(m));
        else
            M = M+1;
            r(M) = r0(idx2(m));
            s(M) = s0(idx2(m));
        end
    end

    % Go through all ranges and find minimum along direction d
    best_fx = inf;
    best_x = nan;
    range_min = -inf;
    gm = -sum(r);
    hm = sum(r .* s);
    for m = 1:M+1
        if m < M+1
            range_max = s(m);
        else
            range_max = inf;
        end
        x = -(q + gm) / p;
        if x < range_min
            x = range_min;
        elseif x > range_max
            x = range_max;
        end
        fx = (0.5 * p * x + (q + gm)) * x + hm;
        if fx < best_fx
            best_fx = fx;
            best_x = x;
        end
        if m < M+1
            range_min = range_max;
            gm = gm + 2 * r(m);
            hm = hm - 2 * r(m) * s(m);
        end
    end
    y0 = y0 + best_x * d;
    Winvy0 = Winvy0 + best_x * Winvd;
    Ay0 = Ay0 + best_x * Ad;
end
y = y0;
